@@ -63,6 +63,7 @@ import java.util.Collections;
 import java.util.Date;
 import java.util.HashMap;
 import java.util.HashSet;
+import java.util.Iterator;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
@@ -214,6 +215,23 @@ public class GoogleDriveFileAccess extends AbstractGoogleDriveAccess implements
                 if (FileStorageFileAccess.NEW != file.getId()) {
                     savedFile.setId(file.getId());
                     if ((null == modifiedFields || modifiedFields.contains(Field.FILENAME)) && false == Strings.isEmpty(file.getFileName())) {
+                        /*
+                         * first check if there is already such a file
+                         */
+                        {
+                            String fileName = file.getFileName();
+                            List<File> hits = searchByFileNamePattern(fileName, file.getFolderId(), false, Arrays.asList(Field.ID, Field.FILENAME), null, null, 0);
+                            boolean found = false;
+                            for (Iterator<File> it = hits.iterator(); !found && it.hasNext();) {
+                                if (it.next().getFileName().equals(fileName)) {
+                                    found = true;
+                                }
+                            }
+
+                            if (found) {
+                                throw FileStorageExceptionCodes.FILE_ALREADY_EXISTS.create();
+                            }
+                        }
                         savedFile.setTitle(file.getFileName());
                         savedFile = drive.files().patch(file.getId(), savedFile).execute();
                     }
@@ -245,7 +263,33 @@ public class GoogleDriveFileAccess extends AbstractGoogleDriveAccess implements
                         }
                     }
                 } else {
-                    savedFile.setTitle(file.getFileName());
+                    /*
+                     * first check if there is already such a file
+                     */
+                    String fileNameToUse = null;
+                    {
+                        List<Field> fields = Arrays.asList(Field.ID, Field.FILENAME);
+                        String name = file.getFileName();
+                        String fileName = name;
+                        int count = 0;
+
+                        while (null == fileNameToUse) {
+                            List<File> hits = searchByFileNamePattern(fileName, file.getFolderId(), false, fields, null, null, 0);
+                            boolean found = false;
+                            for (Iterator<File> it = hits.iterator(); !found && it.hasNext();) {
+                                if (it.next().getFileName().equals(fileName)) {
+                                    found = true;
+                                }
+                            }
+
+                            if (found) {
+                                fileName = FileStorageUtility.enhance(name, ++count);
+                            } else {
+                                fileNameToUse = fileName;
+                            }
+                        }
+                    }
+                    savedFile.setTitle(fileNameToUse);
                     savedFile = drive.files().insert(savedFile).execute();
                 }
                 return new IDTuple(file.getFolderId(), savedFile.getId());
@@ -390,15 +434,52 @@ public class GoogleDriveFileAccess extends AbstractGoogleDriveAccess implements
             }
 
             // Create patch file
+            String title = srcFile.getTitle();
             com.google.api.services.drive.model.File patch = new com.google.api.services.drive.model.File();
-            patch.setParents(Collections.<ParentReference> singletonList(new ParentReference().setId(destFolder)));
+            patch.setParents(Collections.<ParentReference> singletonList(new ParentReference().setId(destId)));
             if (null != update) {
-                if (false == Strings.isEmpty(update.getTitle()) && (null == modifiedFields || modifiedFields.contains(File.Field.FILENAME)) &&
-                    false == update.getTitle().equals(srcFile.getTitle())) {
+                if (false == Strings.isEmpty(update.getTitle()) && (null == modifiedFields || modifiedFields.contains(File.Field.FILENAME)) && false == update.getTitle().equals(srcFile.getTitle())) {
                     patch.setTitle(update.getTitle());
+                    title = update.getTitle();
                 }
             }
 
+            // Check destination folder
+
+            boolean changedName = false;
+            {
+                String baseName;
+                String ext;
+                {
+                    int dotPos = title.lastIndexOf('.');
+                    if (dotPos > 0) {
+                        baseName = title.substring(0, dotPos);
+                        ext = title.substring(dotPos);
+                    } else {
+                        baseName = title;
+                        ext = "";
+                    }
+                }
+                int count = 1;
+                boolean keepOn = true;
+                while (keepOn) {
+                    Drive.Children.List list = drive.children().list(destId);
+                    list.setQ(new StringBuilder().append("title = '").append(title).append("' and ").append(QUERY_STRING_FILES_ONLY_EXCLUDING_TRASH).toString());
+
+                    ChildList childList = list.execute();
+                    if (!childList.getItems().isEmpty()) {
+                        title = new StringBuilder(baseName).append(" (").append(count++).append(')').append(ext).toString();
+                        changedName = true;
+                    } else {
+                        keepOn = false;
+                    }
+                }
+            }
+
+            if (changedName) {
+                patch.setTitle(title);
+            }
+
             // Patch the file
             com.google.api.services.drive.model.File patchedFile = drive.files().patch(id, patch).execute();
 
@@ -1199,11 +1280,14 @@ public class GoogleDriveFileAccess extends AbstractGoogleDriveAccess implements
             com.google.api.services.drive.model.File file = drive.files().get(id).setFields(getFields(fields)).execute();
             checkFileValidity(file);
             List<Revision> revisions = drive.revisions().list(file.getId()).execute().getItems();
-            List<File> versions = new ArrayList<File>(revisions.size());
-            for (int i = 0; i < revisions.size(); i++) {
+            int numRevisions = revisions.size();
+            List<File> versions = new ArrayList<File>(numRevisions);
+            for (int i = 0; i < numRevisions; i++) {
                 Revision revision = revisions.get(i);
-                GoogleDriveFile version = applyRevision(createFile(folderId, id, file, fields), revision, fields, i == revisions.size() - 1);
-                version.setNumberOfVersions(revisions.size());
+                boolean current = i == numRevisions - 1;
+                String name = current ? file.getTitle() : revisions.get(i + 1).getOriginalFilename();
+                GoogleDriveFile version = applyRevision(createFile(folderId, id, file, fields), revision, fields, name, current);
+                version.setNumberOfVersions(numRevisions);
                 versions.add(version);
             }
             /*
@@ -1272,11 +1356,14 @@ public class GoogleDriveFileAccess extends AbstractGoogleDriveAccess implements
                     return metadata;
                 } else {
                     List<Revision> revisions = drive.revisions().list(id).execute().getItems();
-                    metadata.setNumberOfVersions(revisions.size());
-                    for (int i = 0; i < revisions.size(); i++) {
+                    int numRevisions = revisions.size();
+                    metadata.setNumberOfVersions(numRevisions);
+                    for (int i = 0; i < numRevisions; i++) {
                         Revision revision = revisions.get(i);
                         if (revision.getId().equals(version)) {
-                            return applyRevision(metadata, revision, fields, i == revisions.size() - 1);
+                            boolean current = i == numRevisions - 1;
+                            String name = current ? file.getTitle() : revisions.get(i + 1).getOriginalFilename();
+                            return applyRevision(metadata, revision, fields, name, current);
                         }
                     }
                     throw FileStorageExceptionCodes.FILE_VERSION_NOT_FOUND.create(id, version, folderId);
@@ -1328,11 +1415,12 @@ public class GoogleDriveFileAccess extends AbstractGoogleDriveAccess implements
      * @param file The parent file
      * @param revision The revision to apply
      * @param fields The fields to assign, or <code>null</code> to set all fields
+     * @param name The file-name/title to set
      * @param current <code>true</code> if this is the current version, <code>false</code>, otherwise
      * @return The file
      */
-    private GoogleDriveFile applyRevision(GoogleDriveFile file, Revision revision, List<Field> fields, boolean current) throws OXException {
-        file.parseRevision(revision, fields);
+    private GoogleDriveFile applyRevision(GoogleDriveFile file, Revision revision, List<Field> fields, String name, boolean current) throws OXException {
+        file.parseRevision(revision, name, fields);
         file.setIsCurrentVersion(current);
         return file;
     }
